import warnings

from slim_trees import __version__ as slim_trees_version


def check_version(version: str):
    if version != slim_trees_version:
        warnings.warn(
            f"Version mismatch: slim_trees version {slim_trees_version} "
            f"does not match version {version} of the model.",
            stacklevel=2,
        )



import bz2
import gzip
import io
import lzma
import pathlib
import pickle
from collections.abc import Callable
from typing import Any, BinaryIO, Dict, Optional, Tuple, Union


class _NoCompression:
    def __init__(self):
        self.open = open
        self.compress = lambda data: data

    @staticmethod
    def decompress(data):
        return data


def _get_compression_from_path(path: Union[str, pathlib.Path]) -> str:
    compressions = {
        ".gz": "gzip",
        ".lzma": "lzma",
        ".bz2": "bz2",
        ".pickle": "no",
    }
    path = pathlib.Path(path)
    if path.suffix not in compressions:
        raise NotImplementedError(f"File extension '{path.suffix}' not supported")
    return compressions[path.suffix]


def _get_compression_library(compression_method: str) -> Any:
    compression_library = {
        "no": _NoCompression(),
        "lzma": lzma,
        "gzip": gzip,
        "bz2": bz2,
    }
    if compression_method not in compression_library:
        raise ValueError(f"Compression method {compression_method} not implemented.")
    return compression_library[compression_method]


def _get_default_kwargs(compression_method: str) -> Dict[str, Any]:
    defaults = {"gzip": {"compresslevel": 1}}
    fallback: Dict[str, Any] = {}  # else mypy complains
    return defaults.get(compression_method, fallback)


def _unpack_compression_args(
    compression: Optional[Union[str, Dict[str, Any]]] = None,
    file: Optional[Union[str, pathlib.Path, BinaryIO]] = None,
) -> Tuple[str, dict]:
    if compression is not None:
        if isinstance(compression, str):
            return compression, _get_default_kwargs(compression)
        elif isinstance(compression, dict):
            return compression["method"], {
                k: compression[k] for k in compression if k != "method"
            }
        raise ValueError("compression must be either a string or a dict")
    if file is not None and isinstance(file, (str, pathlib.Path)):
        # try to find out the compression using the file extension
        compression_method = _get_compression_from_path(file)
        return compression_method, _get_default_kwargs(compression_method)
    raise ValueError("File must be a path or compression must not be None.")


def dump_compressed(
    obj: Any,
    file: Union[str, pathlib.Path, BinaryIO],
    compression: Optional[Union[str, dict]] = None,
    dump_function: Optional[Callable] = None,
):
    """
    Pickles a model and saves it to the disk. If compression is not specified,
    the compression method will be determined by the file extension.
    :param obj: the object to pickle
    :param file: where to save the object, either a path or a file object
    :param compression: the compression method used. Either a string or a dict with key 'method' set
                        to the compression method and other key-value pairs are forwarded to open()
                        of the compression library.
                        Inspired by the pandas.to_csv interface.
    :param dump_function: the function being called to dump the object, can be a custom pickler.
                          Defaults to pickle.dump()
    """
    if dump_function is None:
        dump_function = pickle.dump

    compression_method, kwargs = _unpack_compression_args(compression, file)
    with _get_compression_library(compression_method).open(
        file, mode="wb", **kwargs
    ) as fd:
        dump_function(obj, fd)


def dumps_compressed(
    obj: Any,
    compression: Optional[Union[str, dict]] = None,
    dump_function: Optional[Callable] = None,
) -> bytes:
    """
    Pickles a model and returns the pickled bytes. If compression is not specified, it won't use
    any.
    :param obj: the object to pickle
    :param compression: the compression method used. Either a string or a dict with key 'method' set
                        to the compression method and other key-value pairs are forwarded to open()
                        of the compression library. Defaults to 'no' compression.
                        Inspired by the pandas.to_csv interface.
    :param dump_function: the function being called to dump the object, can be a custom pickler.
                          Defaults to pickle.dumps()
    """
    if compression is None:
        compression = "no"

    if dump_function is None:
        dump_function = pickle.dumps

    compression_method, kwargs = _unpack_compression_args(compression, None)
    data_uncompressed = dump_function(obj)
    return _get_compression_library(compression_method).compress(data_uncompressed)


def load_compressed(
    file: Union[str, pathlib.Path, BinaryIO],
    compression: Optional[Union[str, dict]] = None,
    unpickler_class: type = pickle.Unpickler,
) -> Any:
    """
    Loads a compressed model.
    :param file: where to load the object from, either a path or a file object
    :param compression: the compression method used. Either a string or a dict with key 'method'
                        set to the compression method and other key-value pairs which are forwarded
                        to open() of the compression library.
                        Inspired by the pandas.to_csv interface.
    :param unpickler_class: custom unpickler class derived from pickle.Unpickler.
                            This is useful to restrict possible imports or to allow unpickling
                            when required module or function names have been refactored.
    """
    compression_method, kwargs = _unpack_compression_args(compression, file)
    with _get_compression_library(compression_method).open(
        file, mode="rb", **kwargs
    ) as fd:
        return unpickler_class(fd).load()


def loads_compressed(
    data: bytes,
    compression: Optional[Union[str, dict]] = None,
    unpickler_class: type = pickle.Unpickler,
) -> Any:
    """
    Loads a compressed model.
    :param data: bytes containing the pickled object.
    :param compression: the compression method used. Either a string or a dict with key 'method'
                        set to the compression method and other key-value pairs which are forwarded
                        to open() of the compression library. Defaults to 'no' compression.
                        Inspired by the pandas.to_csv interface.
    :param unpickler_class: custom unpickler class derived from pickle.Unpickler.
                            This is useful to restrict possible imports or to allow unpickling
                            when required module or function names have been refactored.
    """
    if compression is None:
        compression = "no"

    compression_method, kwargs = _unpack_compression_args(compression, None)
    data_uncompressed = _get_compression_library(compression_method).decompress(data)
    return unpickler_class(io.BytesIO(data_uncompressed)).load()


def get_pickled_size(
    obj: Any,
    compression: Union[str, dict] = "lzma",
    dump_function: Optional[Callable] = None,
) -> int:
    """
    Returns the size that an object would take on disk if pickled.
    :param obj: the object to pickle
    :param compression: the compression method used. Either a string or a dict with key 'method' set
                        to the compression method and other key-value pairs are forwarded to
                        compress() of the compression library.
                        Inspired by the pandas.to_csv interface.
    :param dump_function: the function being called to dump the object, can be a custom pickler.
                          Defaults to pickle.dump()
    """
    compression_method, kwargs = _unpack_compression_args(compression, None)
    if dump_function is None:
        dump_function = pickle.dump

    out = io.BytesIO()
    dump_function(obj, out)
    compressed_bytes = _get_compression_library(compression_method).compress(
        out.getvalue(), **kwargs
    )
    return len(compressed_bytes)



"""
This module changes the default pickle behavior. Instead of using pickle.dump to pickle an object,
you can create a custom Pickler and change the default pickling behavior for certain classes by
changing the dispatch_table of the pickler.
TODO: Update this docstring
In this module, a custom pickling behavior for the sklearn Tree class is specified that compresses
the internal values from int64 to int16 and from float64 to float32 (in the value field) and from
float64 to a mixture between float64 and int16 (in the threshold field).
You can pickle a model with custom picklers by specifying dump_function in dump_compressed in the
pickling module.
"""

import importlib.metadata
import warnings
from pathlib import Path
from typing import Any, BinaryIO, Optional, Union

from slim_trees.pickling import (
    dump_compressed,
    dumps_compressed,
    load_compressed,
    loads_compressed,
)

try:
    __version__ = importlib.metadata.version(__name__)
except Exception as e:
    warnings.warn(f"Could not determine version of {__name__}", stacklevel=1)
    warnings.warn(str(e), stacklevel=1)
    __version__ = "unknown"

__all__ = [
    "dump_compressed",
    "dumps_compressed",
    "load_compressed",
    "loads_compressed",
    "dump_sklearn_compressed",
    "dumps_sklearn_compressed",
    "dump_lgbm_compressed",
    "dumps_lgbm_compressed",
]


def dump_sklearn_compressed(
    model: Any,
    file: Union[str, Path, BinaryIO],
    compression: Optional[Union[str, dict]] = None,
):
    """
    Pickles a model and saves a compressed version to the disk.

    Saves the parameters of the model as int16 and float32 instead of int64 and float64.
    :param model: the model to save
    :param file: where to save the model, either a path or a file object
    :param compression: the compression method used. Either a string or a dict with key 'method' set
                        to the compression method and other key-value pairs are forwarded to `open`
                        of the compression library.
                        Options: ["no", "lzma", "gzip", "bz2"]
    """
    from slim_trees.sklearn_tree import dump

    dump_compressed(model, file, compression, dump)


def dumps_sklearn_compressed(
    model: Any, compression: Optional[Union[str, dict]] = None
) -> bytes:
    """
    Pickles a model and returns the saved object as bytes.

    Saves the parameters of the model as int16 and float32 instead of int64 and float64.
    :param model: the model to save
    :param compression: the compression method used. Either a string or a dict with key 'method' set
                        to the compression method and other key-value pairs are forwarded to `open`
                        of the compression library.
                        Options: ["no", "lzma", "gzip", "bz2"]
    """
    from slim_trees.sklearn_tree import dumps

    return dumps_compressed(model, compression, dumps)


def dump_lgbm_compressed(
    model: Any,
    file: Union[str, Path, BinaryIO],
    compression: Optional[Union[str, dict]] = None,
):
    """
    Pickles a model and saves a compressed version to the disk.

    Saves the parameters of the model as int16 and float32 instead of int64 and float64.
    :param model: the model to save
    :param file: where to save the model, either a path or a file object
    :param compression: the compression method used. Either a string or a dict with key 'method' set
                        to the compression method and other key-value pairs are forwarded to `open`
                        of the compression library.
                        Options: ["no", "lzma", "gzip", "bz2"]
    """
    from slim_trees.lgbm_booster import dump

    dump_compressed(model, file, compression, dump)


def dumps_lgbm_compressed(
    model: Any, compression: Optional[Union[str, dict]] = None
) -> bytes:
    """
    Pickles a model and returns the saved object as bytes.

    Saves the parameters of the model as int16 and float32 instead of int64 and float64.
    :param model: the model to save
    :param compression: the compression method used. Either a string or a dict with key 'method' set
                        to the compression method and other key-value pairs are forwarded to `open`
                        of the compression library.
                        Options: ["no", "lzma", "gzip", "bz2"]
    """
    from slim_trees.lgbm_booster import dumps

    return dumps_compressed(model, compression, dumps)



from typing import Any, Dict

import numpy as np
from numpy.typing import DTypeLike, NDArray

NP_2 = np.__version__ >= "2.0.0"


def can_cast(value: Any, dtype: DTypeLike) -> bool:
    """
    Check if the given value can be cast to the specified data type.

    Notes:
        In numpy 2.0.0, np.can_cast does not support Python scalars anymore and does not apply
          any value-based logic for 0-D arrays and NumPy scalars.
        If value is a uint and dtype is an int of the same size, check if value <= np.iinfo(dtype).max
    """

    if not NP_2:
        return np.can_cast(value, dtype)

    scalar_type = np.min_scalar_type(value)

    return np.can_cast(scalar_type, dtype) or (
        scalar_type.kind == "u"
        and np.issubdtype(dtype, np.signedinteger)
        and value <= np.iinfo(dtype).max
    )


def safe_cast(arr: NDArray, dtype: DTypeLike) -> NDArray:
    if can_cast(arr.max(), dtype) and can_cast(arr.min(), dtype):
        return arr.astype(dtype)
    raise ValueError(f"Cannot cast array to {dtype}.")


def _is_in_neighborhood_of_int(arr: NDArray, iinfo: np.iinfo, eps: float = 1e-12):
    """
    Checks if the numbers are around an integer.
    np.abs(arr % 1 - 1) < eps checks if the number is in an epsilon neighborhood on the right side
    of the next int and arr % 1 < eps checks if the number is in an epsilon neighborhood on the left
    side of the next int.
    """
    return (
        (np.minimum(np.abs(arr % 1 - 1), arr % 1) < eps)
        & (arr >= iinfo.min)
        & (arr <= iinfo.max)
    )


def compress_half_int_float_array(
    a: NDArray, compression_dtype: DTypeLike = "int8"
) -> Dict:
    """Compress small integer and half-integer floats in a lossless fashion

    Idea:
        If most values in array <a> are small integers or half-integers, we can
        store them as float16, while keeping the rest as float64.

    Technical details:
        - The boolean array (2 * a) % 1 == 0 indicates the integers and half-integers in <a>.
        - int8 can represent integers between np.iinfo('int8').min and np.iinfo('int8').max
    """
    info = np.iinfo(compression_dtype)
    a2 = 2.0 * a
    is_compressible = _is_in_neighborhood_of_int(a2, info)

    a2_compressible = a2[is_compressible].astype(compression_dtype)
    a_incompressible = a[~is_compressible]

    state = {
        "is_compressible": np.packbits(is_compressible),
        "a2_compressible": a2_compressible,
        "a_incompressible": a_incompressible,
    }

    return state


def decompress_half_int_float_array(state: Dict) -> NDArray:
    n_thresholds = len(state["a2_compressible"]) + len(state["a_incompressible"])
    is_compressible = np.unpackbits(
        state["is_compressible"], count=n_thresholds
    ).astype("bool")
    a = np.zeros(len(is_compressible), dtype="float64")
    a[is_compressible] = state["a2_compressible"] / 2.0
    a[~is_compressible] = state["a_incompressible"]
    return a



import copyreg
import io
import os
import pickle
import re
import sys
from typing import Any, BinaryIO, Dict, List, Optional, Tuple, Union

import numpy as np
from numpy.typing import DTypeLike, NDArray
from packaging.version import Version

from slim_trees import __version__ as slim_trees_version
from slim_trees.compression_utils import (
    compress_half_int_float_array,
    decompress_half_int_float_array,
    safe_cast,
)
from slim_trees.utils import check_version

try:
    from lightgbm import __version__ as _lightgbm_version
    from lightgbm.basic import Booster

    lightgbm_version = Version(_lightgbm_version)
except ImportError:
    print("LightGBM does not seem to be installed.")
    sys.exit(os.EX_CONFIG)


def dump(model: Any, file: BinaryIO):
    p = pickle.Pickler(file)
    p.dispatch_table = copyreg.dispatch_table.copy()
    p.dispatch_table[Booster] = _booster_pickle
    p.dump(model)


def dumps(model: Any) -> bytes:
    bytes_io = io.BytesIO()
    dump(model, bytes_io)
    return bytes_io.getvalue()


def _booster_pickle(booster: Booster):
    assert isinstance(booster, Booster)
    reconstructor, args, state = booster.__reduce__()  # type: ignore
    compressed_state = _compress_booster_state(state)
    return _booster_unpickle, (
        reconstructor,
        args,
        (slim_trees_version, compressed_state),
    )


def _booster_unpickle(reconstructor, args, compressed_state):
    version, state = compressed_state
    check_version(version)

    booster = reconstructor(*args)
    decompressed_state = _decompress_booster_state(state)
    booster.__setstate__(decompressed_state)
    return booster


_handle_key_name = (
    "_handle" if lightgbm_version.major == 4 else "handle"  # noqa: PLR2004
)


def _compress_booster_state(state: dict):
    """
    For a given state dictionary, store data in a structured format that can then
    be saved to disk in a way that can be compressed.
    """
    assert isinstance(state, dict)
    compressed_state = {k: v for k, v in state.items() if k != _handle_key_name}
    compressed_state["compressed_handle"] = _compress_booster_handle(
        state[_handle_key_name]
    )
    return compressed_state


def _decompress_booster_state(compressed_state: dict):
    assert isinstance(compressed_state, dict)
    state = {k: v for k, v in compressed_state.items() if k != "compressed_handle"}
    state[_handle_key_name] = _decompress_booster_handle(
        compressed_state["compressed_handle"]
    )
    return state


FRONT_STRING_REGEX = r"(?:\w+(?:=.*)?\n)*\n(?=Tree)"
BACK_STRING_REGEX = r"end of trees(?:\n)+(?:.|\n)*"
TREE_GROUP_REGEX = r"(Tree=\d+\n+)((?:.+\n)*)\n\n"

SPLIT_FEATURE_DTYPE = np.int16
THRESHOLD_DTYPE = np.float64
DECISION_TYPE_DTYPE = np.int8
LEFT_CHILD_DTYPE = np.int16
RIGHT_CHILD_DTYPE = LEFT_CHILD_DTYPE
LEAF_VALUE_DTYPE = np.float64


def _extract_feature(feature_line: str) -> Tuple[str, List[str]]:
    feat_name, values_str = feature_line.split("=")
    return feat_name, values_str.split(" ")


def _validate_feature_lengths(feats_map: dict):
    # features on tree-level
    assert len(feats_map["num_leaves"]) == 1
    assert len(feats_map["num_cat"]) == 1
    assert len(feats_map["is_linear"]) == 1
    assert len(feats_map["shrinkage"]) == 1

    # features on node-level
    num_leaves = int(feats_map["num_leaves"][0])
    assert len(feats_map["split_feature"]) == num_leaves - 1
    assert len(feats_map["threshold"]) == num_leaves - 1
    assert len(feats_map["decision_type"]) == num_leaves - 1
    assert len(feats_map["left_child"]) == num_leaves - 1
    assert len(feats_map["right_child"]) == num_leaves - 1

    # features on leaf-level
    num_leaves = int(feats_map["num_leaves"][0])
    assert len(feats_map["leaf_value"]) == num_leaves


def parse(str_list: Union[List[str], List[Optional[str]]], dtype: DTypeLike):
    if np.can_cast(dtype, np.int64):
        int64_array: NDArray = np.array(str_list, dtype=np.int64)
        return safe_cast(int64_array, dtype)
    assert np.can_cast(dtype, np.float64)
    return np.array(str_list, dtype=dtype)


def _compress_booster_handle(model_string: str) -> Tuple[str, List[dict], str]:
    if not model_string.startswith(f"tree\nversion=v{lightgbm_version.major}"):
        raise ValueError(
            f"Only v{lightgbm_version.major} is supported for the booster string format."
        )

    front_str_match = re.search(FRONT_STRING_REGEX, model_string)
    if front_str_match is None:
        raise ValueError("Could not find front string.")

    # delete tree_sizes line since this messes up the tree parsing by LightGBM if not set correctly
    # todo calculate correct tree_sizes
    front_str = re.sub(r"tree_sizes=(?:\d+ )*\d+\n", "", front_str_match.group())

    back_str_match = re.search(BACK_STRING_REGEX, model_string)
    if back_str_match is None:
        raise ValueError("Could not find back string.")
    back_str = back_str_match.group()
    tree_matches = re.findall(TREE_GROUP_REGEX, model_string)
    trees: List[dict] = []
    for i, tree_match in enumerate(tree_matches):
        tree_name, features_list = tree_match
        _, tree_idx = tree_name.replace("\n", "").split("=")
        assert int(tree_idx) == i

        # extract features -- filter out empty ones
        features: List[str] = [f for f in features_list.split("\n") if "=" in f]
        feats_map: Dict[str, List[str]] = dict(_extract_feature(fl) for fl in features)
        _validate_feature_lengths(feats_map)

        tree_values = {
            "num_leaves": int(feats_map["num_leaves"][0]),
            "num_cat": int(feats_map["num_cat"][0]),
            "split_feature": parse(feats_map["split_feature"], SPLIT_FEATURE_DTYPE),
            "threshold": compress_half_int_float_array(
                parse(feats_map["threshold"], THRESHOLD_DTYPE)
            ),
            "decision_type": parse(feats_map["decision_type"], DECISION_TYPE_DTYPE),
            "left_child": parse(feats_map["left_child"], LEFT_CHILD_DTYPE),
            "right_child": parse(feats_map["right_child"], RIGHT_CHILD_DTYPE),
            "leaf_value": parse(feats_map["leaf_value"], LEAF_VALUE_DTYPE),
            "is_linear": int(feats_map["is_linear"][0]),
            "shrinkage": float(feats_map["shrinkage"][0]),
        }

        # if tree is linear, add additional features
        if int(feats_map["is_linear"][0]):
            # attributes: leaf_features, leaf_coeff, leaf_const, num_features
            # TODO: not all of these attributes might be needed.
            tree_values["num_features"] = parse(feats_map["num_features"], np.int32)
            tree_values["leaf_const"] = parse(feats_map["leaf_const"], LEAF_VALUE_DTYPE)
            tree_values["leaf_features"] = parse(
                [s if s else "-1" for s in feats_map["leaf_features"]],
                np.int16,
            )
            tree_values["leaf_coeff"] = parse(
                [s if s else None for s in feats_map["leaf_coeff"]], np.float64
            )

        # at last
        trees.append(tree_values)

    return front_str, trees, back_str


def _validate_tree_structure(tree: dict):
    assert isinstance(tree, dict)
    if tree.keys() != {
        "num_leaves",
        "num_cat",
        "split_feature",
        "threshold",
        "decision_type",
        "left_child",
        "right_child",
        "leaf_value",
        "is_linear",
        "shrinkage",
    }:
        raise ValueError(
            "Invalid tree structure. Do you use an unsupported LightGBM version or try to load a "
            "model that was pickled with a different version of LightGBM?"
        )


def _decompress_booster_handle(compressed_state: Tuple[str, List[dict], str]) -> str:
    front_str, trees, back_str = compressed_state
    assert isinstance(front_str, str)
    assert isinstance(trees, list)
    assert isinstance(back_str, str)

    handle = front_str
    for i, tree in enumerate(trees):
        _validate_tree_structure(tree)
        is_linear = int(tree["is_linear"])

        num_leaves = len(tree["leaf_value"])
        num_nodes = len(tree["split_feature"])

        tree_str = f"Tree={i}\n"
        tree_str += f"num_leaves={tree['num_leaves']}\nnum_cat={tree['num_cat']}\nsplit_feature="
        tree_str += " ".join([str(x) for x in tree["split_feature"]])
        tree_str += "\nsplit_gain=" + ("0 " * num_nodes)[:-1]
        threshold = decompress_half_int_float_array(tree["threshold"])
        tree_str += "\nthreshold=" + " ".join([str(x) for x in threshold])
        tree_str += "\ndecision_type=" + " ".join(
            [str(x) for x in tree["decision_type"]]
        )
        tree_str += "\nleft_child=" + " ".join([str(x) for x in tree["left_child"]])
        tree_str += "\nright_child=" + " ".join([str(x) for x in tree["right_child"]])
        tree_str += "\nleaf_value=" + " ".join([str(x) for x in tree["leaf_value"]])
        tree_str += "\nleaf_weight=" + ("0 " * num_leaves)[:-1]
        tree_str += "\nleaf_count=" + ("0 " * num_leaves)[:-1]
        tree_str += "\ninternal_value=" + ("0 " * num_nodes)[:-1]
        tree_str += "\ninternal_weight=" + ("0 " * num_nodes)[:-1]
        tree_str += "\ninternal_count=" + ("0 " * num_nodes)[:-1]
        tree_str += f"\nis_linear={tree['is_linear']}"
        if is_linear:
            tree_str += "\nleaf_const=" + " ".join(str(x) for x in tree["leaf_const"])
            tree_str += "\nnum_features=" + " ".join(
                str(x) for x in tree["num_features"]
            )
            tree_str += "\nleaf_features=" + " ".join(
                "" if f == -1 else str(int(f)) for f in tree["leaf_features"]
            )
            tree_str += "\nleaf_coeff=" + " ".join(
                "" if np.isnan(f) else str(f) for f in tree["leaf_coeff"]
            )

        tree_str += f"\nshrinkage={int(tree['shrinkage'])}"

        tree_str += "\n\n\n"
        handle += tree_str

    handle += back_str
    return handle



import io
import os
import sys

from packaging.version import Version

from slim_trees import __version__ as slim_trees_version
from slim_trees.compression_utils import (
    compress_half_int_float_array,
    decompress_half_int_float_array,
    safe_cast,
)
from slim_trees.utils import check_version

try:
    from sklearn import __version__ as _sklearn_version
    from sklearn.tree._tree import Tree

    sklearn_version = Version(_sklearn_version)
    sklearn_version_ge_130 = sklearn_version >= Version("1.3")
except ImportError:
    print("scikit-learn does not seem to be installed.")
    sys.exit(os.EX_CONFIG)

import copyreg
import pickle
from typing import Any, BinaryIO, Dict

import numpy as np
from numpy.typing import NDArray


def dump(model: Any, file: BinaryIO):
    p = pickle.Pickler(file)
    p.dispatch_table = copyreg.dispatch_table.copy()
    p.dispatch_table[Tree] = _tree_pickle
    p.dump(model)


def dumps(model: Any) -> bytes:
    bytes_io = io.BytesIO()
    dump(model, bytes_io)
    return bytes_io.getvalue()


def _tree_pickle(tree: Tree):
    assert isinstance(tree, Tree)
    reconstructor, args, state = tree.__reduce__()
    compressed_state = _compress_tree_state(state)  # type: ignore
    return _tree_unpickle, (reconstructor, args, (slim_trees_version, compressed_state))


def _tree_unpickle(reconstructor, args, compressed_state):
    version, state = compressed_state
    check_version(version)

    tree = reconstructor(*args)
    decompressed_state = _decompress_tree_state(state)
    tree.__setstate__(decompressed_state)
    return tree


def _compress_tree_state(state: Dict) -> Dict:
    """
    Compresses a Tree state.
    :param state: dictionary with 'max_depth', 'node_count', 'nodes', 'values' as keys.
    :return: dictionary with compressed tree state, only with data that is relevant for prediction.
    """
    assert isinstance(state, dict)
    assert state.keys() == {"max_depth", "node_count", "nodes", "values"}
    nodes = state["nodes"]
    # nodes is a numpy array of tuples of the following form
    # (left_child, right_child, feature, threshold, impurity, n_node_samples,
    #  weighted_n_node_samples)
    dtype_child = np.uint16
    dtype_feature = np.uint16
    dtype_threshold = np.float64
    dtype_value = np.float32

    children_left = nodes["left_child"]
    children_right = nodes["right_child"]

    is_leaf = children_left == -1
    # assert that the leaves are the same no matter if you use children_left or children_right
    assert np.array_equal(is_leaf, children_right == -1)

    # feature, threshold and children are irrelevant when leaf

    children_left = safe_cast(children_left[~is_leaf], dtype_child)
    children_right = safe_cast(children_right[~is_leaf], dtype_child)
    features = safe_cast(nodes["feature"][~is_leaf], dtype_feature)
    # value is irrelevant when node not a leaf
    values = safe_cast(state["values"][is_leaf], dtype_value)
    # do lossless compression for thresholds by downcasting half ints (e.g. 5.5, 10.5, ...) to int8
    thresholds = nodes["threshold"][~is_leaf].astype(dtype_threshold)
    thresholds = compress_half_int_float_array(thresholds)

    if sklearn_version_ge_130:
        missing_go_to_left = nodes["missing_go_to_left"][~is_leaf].astype("bool")
    else:
        missing_go_to_left = None

    # TODO: make prettier once python 3.8 is not supported anymore
    return {
        **{
            "max_depth": state["max_depth"],
            "node_count": state["node_count"],
            "is_leaf": np.packbits(is_leaf),
            "children_left": children_left,
            "children_right": children_right,
            "features": features,
            "thresholds": thresholds,
            "values": values,
        },
        **(
            {"missing_go_to_left": np.packbits(missing_go_to_left)}  # type: ignore
            if sklearn_version_ge_130
            else {}
        ),
    }


def _decompress_tree_state(state: Dict) -> Dict:
    """
    Decompresses a Tree state.
    :param state: 'children_left', 'children_right', 'features', 'thresholds', 'values' as keys.
                  If the sklearn version is >=1.3.0, also 'missing_go_to_left' is a key.
                  'max_depth' and 'node_count' are passed through.
    :return: dictionary with decompressed tree state.
    """
    assert isinstance(state, dict)
    # TODO: make prettier once python 3.8 is not supported anymore
    if state.keys() != {
        *{
            "max_depth",
            "node_count",
            "is_leaf",
            "children_left",
            "children_right",
            "features",
            "thresholds",
            "values",
        },
        *({"missing_go_to_left"} if sklearn_version >= Version("1.3") else set()),
    }:
        raise ValueError(
            "Invalid tree structure. Do you use an unsupported scikit-learn version "
            "or try to load a model that was pickled with a different version of scikit-learn?"
        )
    n_nodes = state["node_count"]

    children_left: NDArray = np.zeros(n_nodes, dtype=np.int64)
    children_right: NDArray = np.zeros(n_nodes, dtype=np.int64)
    features: NDArray = np.zeros(n_nodes, dtype=np.int64)
    thresholds: NDArray = np.zeros(n_nodes, dtype=np.float64)
    # same shape as values but with all nodes instead of only the leaves
    values: NDArray = np.zeros((n_nodes, *state["values"].shape[1:]), dtype=np.float64)
    missing_go_to_left: NDArray = np.zeros(n_nodes, dtype="uint8")

    is_leaf = np.unpackbits(state["is_leaf"], count=n_nodes).astype("bool")
    children_left[~is_leaf] = state["children_left"]
    children_left[is_leaf] = -1
    children_right[~is_leaf] = state["children_right"]
    children_right[is_leaf] = -1
    features[~is_leaf] = state["features"]
    features[is_leaf] = -2  # feature of leaves is -2
    thresholds[~is_leaf] = decompress_half_int_float_array(state["thresholds"])
    thresholds[is_leaf] = -2.0  # threshold of leaves is -2
    values[is_leaf] = state["values"]
    if sklearn_version_ge_130:
        missing_go_to_left[~is_leaf] = np.unpackbits(
            state["missing_go_to_left"], count=(~is_leaf).sum()
        )

    dtype = np.dtype(
        [
            ("left_child", "<i8"),
            ("right_child", "<i8"),
            ("feature", "<i8"),
            ("threshold", "<f8"),
            ("impurity", "<f8"),
            ("n_node_samples", "<i8"),
            ("weighted_n_node_samples", "<f8"),
        ]
        + ([("missing_go_to_left", "<u1")] if sklearn_version_ge_130 else [])
    )
    nodes = np.zeros(n_nodes, dtype=dtype)
    nodes["left_child"] = children_left
    nodes["right_child"] = children_right
    nodes["feature"] = features
    nodes["threshold"] = thresholds
    if sklearn_version_ge_130:
        nodes["missing_go_to_left"] = missing_go_to_left

    return {
        "max_depth": state["max_depth"],
        "node_count": state["node_count"],
        "nodes": nodes,
        "values": values,
    }
